#!/usr/bin/env python
# coding: utf-8

# ## TUTORATO 3  : HEAP 
# Heapsort introduce una tecnica di progetto di algoritmi: si utilizza una struttura di dati, denominata _heap_, per gestire le informazioni durante l'esecuzione dell'algoritmo.

# La struttura di dati **heap** è un array che può essere visto come un albero binario completo.
# - Ogni nodo dell'albero corrisponde a un elemento dell'array che contiene il valore del nodo. 
# - L'albero è riempito su tutti i livelli tranne, eventualmente, il più basso che è riempito da sinistra in poi.
# - Gli _heap_ soddisfano la _proprietà di ordinamento parziale delle heap_: il valore di un nodo è minore uguale al valore del padre: quindi l'elemento più grande è memorizzato nella radice e is ottoalberi di qualunque nodo contengono valori più piccoli del nodo stesso.

# Come porcederemo per implementare questa struttura dati:
# - Implementeremo una classe BinHeap() che sarà la classe delle Heap Binarie
# - Implementeremo il costruttore della classe, che conterrà i parametri necessari per definire una Heap a partire da un array
# - Implementeremo una funzione che dato l'indice del nodo genitore con indice $i$ all'interno dell'array ne ritorna il figlio sinistro: left($i$)
# - Implementeremo una funzione che dato l'indice del nodo genitore con indice $i$ all'interno dell'array ne ritorna il figlio destro: right($i$)
# - Implementeremo la funzione Heapify($i$). Questa funzione assume che gli alberi binari con radice left($i$) e right($i$) siano alberi heap e lascia scendere il nodo con indice $i$ nella posizione corretta affinche venga rispettata la _proprietà di ordinamento parziale delle heap_
# - Implementeremo la funzione che costruisce le Heap a partire dall'array, utilizzando la funzione Heapify() appena definita

# In[4]:


class BinHeap():
    
    def __init__(self, list_keys): # costruttore 
        self._keys = list_keys 
        self.tree_dimension = len(list_keys)
        
    def left(self, i): # dato l'indice di un nodo torna l'indice del nodo del figlio sx 
        return 2*i + 1
        
    def right(self, i):
        return 2*i + 2
        
    def heapify(self, i):
        L = self.left(i)
        R = self.right(i)
        largest = i 
        if L < self.tree_dimension and self._keys[L] > self._keys[largest] :
            largest = L 
        if  R < self.tree_dimension and self._keys[R] > self._keys[largest]:
            largest = R
        if largest != i :
            tmp = self._keys[i]
            self._keys[i] = self._keys[largest]
            self._keys[largest] = tmp 
            self._keys = self.heapify(largest) 
        return self._keys
        
        
    def build_heap(self):
        '''
        poichè tutti gli elementi del sottoarray dopo n/2 + 1 sono foglie dell'albero, ognuna è una heap
        di un solo elemento.
        La procedure build_heap attraversa i restanti nodi dell'albero ed esegue Heapify su ognuno di essi.
        '''
        for i in range(self.tree_dimension // 2 , - 1, -1):
            self.heapify(i)
        return self._keys
        
        
    def heapsort(self):
        '''
        L'algoritmo comincia costruendo una heap sull'array di input. 
        L'elemento più grande dell'array è posto nella radice quindi può essere posto nella corretta posizione finale. 
        Posso facilmente trasformare la lista rimanente in una Heap, infatti:
        - I figli della radice rimangono heap 
        - L'unico elelmento che può violare la proprietà di ordinamento parziale delle heap è la radice
        Applico Heapify() al primo elemento e lo metto nella posizione corretta. 
        '''
        pass
    
        
    def __repr__(self):
        bt_str = '['

        level = 0
        next_level = 2 ** level
        for i in range(0, self.tree_dimension):
            bt_str += '{}'.format(self._keys[i])
            if i + 1 < self.tree_dimension:
                if next_level == i + 1:
                    bt_str += ']\n['
                else:
                    bt_str += ' '
            if next_level == i + 1:
                level += 1
                next_level += 2 ** level

        return bt_str + ']'


# In[5]:


A = [1, 2, 38, 7, 4, 7, 88, 74, 32, 89, 56, 6, 31, 100, 130, 0]


# In[6]:


a = BinHeap(A)
print(a)


# In[12]:


b = BinHeap(A)
# print(b)


# In[7]:


Aord = a.build_heap()
print(Aord)


# In[8]:


B = b.heapsort()
# print(B)


# In[ ]:




